"""DeCoHde snake edition
v01.0
23.04.2020"""



import math
import random
import copy


class Vehicle:
    def __init__(self, name, manpower, fuel, pop, armor, hitpoints, rear_armor, target_size, speed, acc_far, acc_mid,
                 acc_near, damage, deflection_dmg_modifier, pen_far, pen_mid, pen_near, range, sight, reload_min,
                 reload_max, dis_far, dis_mid, dis_near, scatter_offset, scatter_ratio, distance_scatter_max,
                 scatter_angle, hitbox_length, hitbox_broadness, ready_aim_time, wind_up, wind_down, fire_aim_time):
        self.name = name
        self.Manpower = manpower
        self.fuel = fuel
        self.pop = pop
        self.armor = armor
        self.hitpoints = hitpoints
        self.rear_armor = rear_armor
        self.target_size = target_size
        self.speed = speed
        self.acc_far = acc_far
        self.acc_mid = acc_mid
        self.acc_near = acc_near
        self.damage = damage
        self.deflection_damage_modifier = deflection_dmg_modifier
        self.pen_far = pen_far
        self.pen_mid = pen_mid
        self.pen_near = pen_near
        self.range = range
        self.sight = sight
        self.reload_min = reload_min
        self.reload_max = reload_max
        self.dis_far = dis_far
        self.dis_mid = dis_mid
        self.dis_near = dis_near
        self.scatter_offset = scatter_offset
        self.scatter_ratio = scatter_ratio
        self.distance_scatter_max = distance_scatter_max
        self.scatter_angle = scatter_angle
        self.hitbox_length = hitbox_length
        self.hitbox_broadness = hitbox_broadness
        self.ready_aim_time = ready_aim_time
        self.wind_up = wind_up
        self.wind_down = wind_down
        self.fire_aim_time = fire_aim_time

        self.orientation = ''
        self.edge_points = None
        self.time_to_next_shot = 0
        self.shot_counter = 0
        self.scatter_shot_counter = 0
        self.natural_hit_counter = 0
        self.hit_counter = 0
        self.miss_counter = 0
        self.pen_counter = 0
        self.bounce_counter = 0
        self.frontal_hit_counter = 0
        self.rear_hit_counter = 0
        self.average_time_to_shoot = 0

        self.calc_avg_time_to_next_shot()


    def calc_pen(self, distance):
        """Calculates the penetration at given range. Logic: Takes penetration of the neighbouring reference points
                and weights them according to the actual range"""

        if distance > self.range:
            raise Exception("Distance too long for selected vehicle")

        penetration = None

        if distance > self.range:
            penetration = 0
        elif distance <= self.dis_near:
            penetration = self.pen_near
        elif distance <= self.dis_mid:
            penetration = self.pen_near * (
                        1 - (distance - self.dis_near) / (self.dis_mid - self.dis_near)) + self.pen_mid * (
                                      (distance - self.dis_near) / (self.dis_mid - self.dis_near))
        elif distance <= self.dis_far:
            penetration = self.pen_mid * (1 - (distance - self.dis_mid) / (self.dis_far - self.dis_mid)) + self.pen_far * (
                        (distance - self.dis_mid) / (self.dis_far - self.dis_mid))
        elif distance <= self.range:
            penetration = self.pen_far
        return penetration


    def calc_base_acc(self, distance):
        """Calculates the base accuracy at given range. Must me multiplied with the target size. Logic:
                Takes accuracy of the neighbouring reference points and weights them according to the actual range"""

        if distance > self.range:
            raise Exception("Distance too long for selected vehicle")

        base_accuracy = None

        if distance > self.range:
            base_accuracy = 0
        elif distance <= self.dis_near:
            base_accuracy = self.acc_near
        elif distance <= self.dis_mid:
            base_accuracy = self.acc_near * (
                        1 - (distance - self.acc_near) / (self.acc_mid - self.dis_near)) + self.acc_mid * (
                                    (distance - self.dis_near) / (self.dis_mid - self.dis_near))
        elif distance <= self.dis_far:
            base_accuracy = self.acc_mid * (
                         1 - (distance - self.dis_mid) / (self.dis_far - self.dis_mid)) + self.acc_far * (
                                    (distance - self.dis_mid) / (self.dis_far - self.dis_mid))
        elif distance <= self.range:
            base_accuracy = self.acc_far
        return base_accuracy


    def calc_time_to_next_shot(self, delay=0):
        """Calculates the time it takes to the next shot based on the ROF stats. Delay is important for the first shot
        only."""

        tbs_correction_value = 0

        for i in self.wind_down, self.wind_up, self.fire_aim_time, self.reload_max:
            tbs_correction_value += 0.125 if i != 0 else 0

        self.time_to_next_shot = self.wind_up + self.wind_down + self.fire_aim_time + \
                            random.uniform(self.reload_min, self.reload_max) + tbs_correction_value + delay


    def calc_avg_time_to_next_shot(self):
        """Calculates the AVERAGE time it takes to shoot based on the ROF stats. Only important for the inspector."""

        tbs_correction_value = 0

        for i in self.wind_down, self.wind_up, self.fire_aim_time, self.reload_max:
            tbs_correction_value += 0.125 if i != 0 else 0

        self.average_time_to_shoot = self.wind_up + self.wind_down + self.fire_aim_time + \
                            (self.reload_min + self.reload_max) / 2 + tbs_correction_value


def build_unit_dictionary():
    unit_dic = {
                "StuGIII": Vehicle('StugIII Ausf. G', 280, 90, 10, 140, 560, 70, 17, 6, 0.04, 0.05, 0.06, 160, 0, 170,
                                   185, 200, 50, 35, 4.5, 5.5, 50, 25, 0, 0.25, 1, 4, 5, 5.56, 2.9, 0.125, 0, 0, 0.125),
                "OST_P4": Vehicle('OST P4', 350, 120, 12, 180, 640, 90, 20, 6.3, 0.025, 0.0375, 0.05, 160, 0, 110, 115,
                                  125, 40, 35, 5.3, 5.7, 40, 20, 0, 0.25, 1, 6.4, 7.2, 10, 3.4, 0.125, 0, 0, 0.125),
                "OST_Panther": Vehicle('OST Panther', 490, 185, 18, 260, 960, 90, 24, 6.6, 0.035, 0.045, 0.06, 160, 0,
                                       220, 240, 260, 50, 35, 5.2, 5.6, 50, 25, 0,0.25, 1, 8, 7.5,6.87, 3.27,0.125, 0,
                                       1, 0.125),
                "OST_Tiger": Vehicle('OST Tiger', 640, 230, 21, 300, 1040, 140, 26, 5.2, 0.025, 0.0375, 0.05, 160,  0,
                                     200, 220, 240, 45, 35, 4.7, 5.3, 45, 22.5, 0, 0.25, 1, 4.3, 4.75, 8.45, 3.7, 0.125,
                                     0, 0, 0.125),
                "OST_Tiger_vet3": Vehicle('OST_Tiger_vet3', 640, 230, 21, 300, 1040, 140, 26, 5.2, 0.03, 0.045, 0.06,
                                          160, 0, 200, 220, 240, 50, 35, 3.29, 3.71, 50, 25, 0, 0.25, 0.8, 3.44, 3.8,
                                          8.45, 3.7, 0.125, 0, 0, 0.125),
                "Elefant": Vehicle('Elefant', 720, 245, 22, 400, 1040, 110, 26, 3.3, 0.03, 0.45, 0.55, 300, 0, 360,
                                   400, 440, 70, 35, 8, 9, 70, 35, 0, 0.25, 1, 10, 5, 8.14, 3.38, 0.125, 0, 0, 0.125),


                "T-70": Vehicle('T-70', 260, 70, 6, 70, 400, 35, 18, 6.9, 0.025, 0.0375, 0.05, 40,  0, 40, 45, 50, 40,
                                35, 2.1, 2.5, 40, 20, 0,0.15, 1, 2.5, 7.5,4.29, 2.32,0.125, 0, 0, 0.125),
                "SU-76": Vehicle('SU76', 280, 75, 8, 70, 400, 35, 20, 6.3, 0.0375, 0.04, 0.05, 120,  0, 160, 170, 180,
                                 60, 35, 3.9, 4.4, 60, 30, 0,0, 0.47, 4.7, 5,4.88, 2.73,0.125, 0, 0.875, 0.125),
                "T34/76": Vehicle('T34/76', 300, 90, 10, 150, 640, 75, 20, 6.5, 0.025, 0.0375, 0.05, 160,  0, 80, 100,
                                  120, 40, 35, 6.1, 6.1, 40, 20, 0, 0.25, 1, 6.9, 7.5, 6.68, 3, 0.125, 0, 0, 0.125),
                "T34/85": Vehicle('T34/85', 380, 130, 14, 160, 800, 80, 22, 6.3, 0.025, 0.0375, 0.05, 160, 0, 120, 140,
                                  160, 40, 35, 6, 6.3, 40, 20, 0,0.25, 1, 5, 7.5,6.68, 3,0.125, 0, 0, 0.125),
                "KV-1": Vehicle('KV-1', 420, 145, 14, 270, 800, 165, 24, 5.1, 0.025, 0.0375, 0.05, 160, 0, 80, 100, 120,
                                40, 35, 5.2, 5.8, 40, 20, 0,0.25, 1, 6.9, 7.5,8.45, 3.25,0.125, 0, 0, 0.125),
                "SU-85": Vehicle('SU-85', 350, 130, 15, 140, 640, 70, 18, 5.7, 0.04, 0.045, 0.055, 160, 0, 220, 230,
                                 240, 60, 35/75, 5.4, 5.4, 60, 30, 0,0.25, 1, 7.5, 5,6.10, 3,0.125, 0, 0, 0.125),
                "IS-2": Vehicle('IS-2', 640, 230, 21, 340, 1040, 140, 26, 5, 0.025, 0.0375, 0.05, 160,  0, 210, 230,
                                250, 40, 35, 6.2, 6.6, 40, 20, 0, 0.185, 1, 5.7, 7.5, 9.9, 3.09, 0.125, 0, 0, 0.125),
                "ISU-152_AP": Vehicle('ISU-152(AP)', 680, 260, 22, 340, 1040, 155, 26, 4.5, 0.03, 0.0425, 0.055, 240,
                                      0, 200, 230, 260, 70, 35, 9, 11, 70, 35, 0, 0.5, 0.087, 8.7, 5, 9.18, 3.07, 0.125,
                                      0, 0, 0.125),
                "KV-2": Vehicle('KV2', 630, 230, 21, 300, 1040, 120, 26, 4.9, 0.025, 0.0375, 0.05, 240, 0, 140, 150,
                                170, 40, 35, 9, 9, 40, 20, 0, 0.25, 0.185, 7.4, 5, 6.67, 3.35, 0.125, 0, 0, 0.125),


                "Puma": Vehicle('Puma', 320, 70, 7, 25, 400, 12.5, 18, 7.2, 0.025, 0.0375, 0.05, 120, 80, 120, 160, 0,
                                50, 50, 3.6, 4.1, 50, 25, 0, 0.25, 1, 9.5, 7.5, 6.02, 2.36, 0.125, 0, 0, 0.125),
                "OKW_P4": Vehicle('OKW P4', 380, 140, 14, 234, 640, 90, 22, 6, 0.025, 0.0375, 0.05, 160,  0, 110, 115,
                                  125, 40, 35, 5.3, 5.7, 40, 20, 0,0.25, 1, 5.54, 6.5,7.02, 2.88,0.125, 0, 0, 0.125),
                "JagdpanzerIV": Vehicle('JagdpanzerIV', 400, 135, 15, 230, 640, 80, 17, 5.5, 0.04, 0.05, 0.06, 160, 0,
                                        170, 185, 200, 60, 45, 4.3, 5.2, 60, 30, 0,0.25, 1, 8, 7.5,6.85, 3.17,0.125, 0,
                                        0, 0.125),
                "OKW_Panther": Vehicle('OKW Panther', 490, 185, 18, 260, 960, 90, 24, 6.6, 0.035, 0.045, 0.06, 160, 0,
                                       220, 240, 260, 50, 35, 5.2, 5.6, 50, 25, 0,0.25, 1, 8, 7.5,6.87, 3.27,0.125, 0,
                                       1, 0.125),
                "King Tiger": Vehicle('King Tiger', 720, 280, 23, 375, 1280, 150, 26, 3.8, 0.025, 0.0375, 0.05, 240, 0,
                                      200, 220, 240, 45, 35, 5.7, 6.3, 45, 22.5, 0, 0.185, 1, 5.7, 5, 7.38, 3.755,
                                      0.125, 0, 0, 0.125),
                "Jagdtiger": Vehicle('Jagdtiger', 720, 245, 23, 525, 1280, 110, 26, 3, 0.03, 0.045, 0.055, 300, 500,  0,
                                     525, 550, 80, 35, 8.55, 8.55, 80, 40, 0, 0.06, 1, 6.8, 7.5, 10.65, 3.6, 0.125, 0,
                                     0,
                                     0.125),


                "Stuart": Vehicle('Stuart', 270, 70, 6, 80, 400, 50, 19, 7, 0.025, 0.0375, 0.05, 80, 0, 55, 60, 75, 40,
                                  35, 2.9, 4.4, 40, 20, 0,0.15, 1, 4.2, 5,4.84, 2.29,1.5, 0, 0, 0.125),
                "M4A3 Sherman": Vehicle('M4A3 Sherman', 340, 110, 12, 160, 640, 80, 21, 6.4, 0.025, 0.0375, 0.05, 160,
                                        0, 100, 120, 140, 40, 35, 5.6, 6, 40, 20, 0, 0.3, 1, 6.2, 6, 5.85, 2.74, 0.125,
                                        0, 0, 0.125),
                "Sherman(76mm)": Vehicle('Sherman(76mm)', 380, 135, 14, 160, 640, 80, 23, 6.5, 0.025, 0.0375, 0.05, 160,
                                         0, 120, 130, 140, 45, 35, 4.1, 4.5, 45, 22.5, 0,0.25, 1, 5, 7.5, 5.85, 2.74,
                                         0.125, 0, 0, 0.125),
                "Sherman E8": Vehicle('Sherman Easy8', 380, 140, 14, 215, 720, 95, 23, 6.1, 0.025, 0.0375, 0.05, 160, 0,
                                      155, 165, 200, 40, 35, 6, 6.6, 40, 20, 0,0.325, 1, 5.4, 5,5.85, 2.74,0.125, 0, 0,
                                      0.125),
                "Jackson": Vehicle('M36 Jackson', 400, 145, 16, 130, 640, 60, 24, 6.5, 0.035, 0.045, 0.05, 160, 0, 220,
                                   240, 260, 60, 35, 4.375, 4.975, 60, 30, 0,0.25, 1, 8, 5,5.97, 3.05,0.125, 0.5, 1.125,
                                   0.125),
                "Pershing": Vehicle('Pershing', 640, 230, 19, 270, 960, 110, 24, 6, 0.035, 0.045, 0.06, 160, 0, 220,
                                    240, 260, 45, 35, 5.5, 5.5, 45, 22.5, 0, 0.3, 1, 4.2, 6, 6.337, 3.51, 0.125, 0, 1,
                                    0.125),


                "AEC": Vehicle('AEC', 280, 60, 6, 55, 400, 12.5, 18, 7.2, 0.03125, 0.0375, 0.05, 120, 0, 80, 100, 120,
                               40, 50, 3.6, 4.1, 40, 20, 0,0.25, 1, 9.5, 7.5,5.18, 2.74,0.125, 0, 0, 0.125),
                "Valentine": Vehicle('Valentine', 300, 80, 7, 120, 480, 80, 18, 7, 0.025, 0.0375, 0.05, 120, 0, 80, 100,
                                     120, 40, 35, 5, 5, 40, 20, 0,0.3, 1, 6.2, 6,5.41, 2.629,0.125, 0, 0, 0.125),
                "Cromwell": Vehicle('Cromwell', 340, 110, 12, 160, 640, 80, 20, 7, 0.025, 0.0375, 0.05, 160, 0, 105,
                                    120, 135, 40, 35, 5.6, 6, 40, 20, 0,0.25, 1, 6.4, 7.5,6.35, 2.908,0.125, 0, 0,
                                    0.125),
                "Firefly": Vehicle('Firefly', 440, 145, 16, 160, 640, 95, 23, 5.3, 0.04, 0.05, 0.06, 200, 0, 210, 240,
                                   260, 60, 35, 8, 8, 60, 30, 0,0.25, 1, 8, 7.5,5.89, 2.64,0.125, 0, 0, 0.125),
                "Comet": Vehicle('Comet', 480, 185, 18, 290, 800, 110, 22, 6.9, 0.03, 0.045, 0.06, 160, 0, 170, 190,
                                 210, 45, 35, 5.9, 6.4, 45, 22.5, 0, 0.25, 1, 6.4, 5.25, 7.661, 3.07, 0.125, 0, 0,
                                 0.125),
                }

    return unit_dic


def translocate_vehicle(vehicle, distance, rotation_angle):
    """Calculates the coordinates of a tanks the edges of a tanks hitbox and translocates them to the set distance.
    Also reports orientation in the form of the most forward edge"""
    # TODO: Neu formatieren um Daten von Target automatisch zu kriegen; Winkel darf max 90° sein

    length = vehicle.hitbox_length
    broadness = vehicle.hitbox_broadness

    rad_rotation_angle = math.radians(rotation_angle)

    front_left = (-length / 2 * math.cos(rad_rotation_angle) - broadness / 2 * math.sin(rad_rotation_angle),
                  length / 2 * math.sin(rad_rotation_angle) - broadness / 2 * math.cos(rad_rotation_angle), "FL")
    front_right = (-length / 2 * math.cos(rad_rotation_angle) + broadness / 2 * math.sin(rad_rotation_angle),
                   length / 2 * math.sin(rad_rotation_angle) + broadness / 2 * math.cos(rad_rotation_angle), "FR")
    back_left = (length / 2 * math.cos(rad_rotation_angle) - broadness / 2 * math.sin(rad_rotation_angle),
                 -length / 2 * math.sin(rad_rotation_angle) - broadness / 2 * math.cos(rad_rotation_angle), "BL")
    back_right = (length / 2 * math.cos(rad_rotation_angle) + broadness / 2 * math.sin(rad_rotation_angle),
                  -length / 2 * math.sin(rad_rotation_angle) + broadness / 2 * math.cos(rad_rotation_angle), "BR")

    if rotation_angle == 0 or rotation_angle == 90:
        transloc_front_left = (front_left[0] - front_left[0] + distance, front_left[1])
        transloc_front_right = (front_right[0] - front_left[0] + distance, front_right[1])
        transloc_back_left = (back_left[0] - front_left[0] + distance, back_left[1])
        transloc_back_right = (back_right[0] - front_left[0] + distance, back_right[1])

        orientation = 'FL'

    elif rotation_angle == 180 or rotation_angle == 270:
        transloc_front_left = (front_left[0] - back_right[0] + distance, front_left[1])
        transloc_front_right = (front_right[0] - back_right[0] + distance, front_right[1])
        transloc_back_left = (back_left[0] - back_right[0] + distance, back_left[1])
        transloc_back_right = (back_right[0] - back_right[0] + distance, back_right[1])

        orientation = 'BR'

    else:
        # find the three points closest to the shooter (first three after sorting ascending x-Value). First value
        # is also the middle one on the Y-axis. Draw lines to the neighbouring edges, make IP with x-Axis. The IP with
        # with greater x is closer to 0/0 and all edges get translocated with this x value.

        edge_points = (front_right, front_left, back_left, back_right)
        sorted_edge_points = sorted(edge_points, key=lambda point: point[0])  # find the point most to the left
        del sorted_edge_points[3]

        orientation = sorted_edge_points[0][2]

        if orientation == 'FL':
            top_m, top_b = find_line_through_2_points(front_left, front_right)
            top_IP_x = -top_b / top_m
            bottom_m, bottom_b = find_line_through_2_points(front_left, back_left)
            bottom_IP_x = -bottom_b / bottom_m

        elif orientation == 'BL':
            top_m, top_b = find_line_through_2_points(back_left, front_left)
            top_IP_x = -top_b / top_m
            bottom_m, bottom_b = find_line_through_2_points(back_left, back_right)
            bottom_IP_x = -bottom_b / bottom_m

        elif orientation == 'BR':
            top_m, top_b = find_line_through_2_points(back_right, back_left)
            top_IP_x = -top_b / top_m
            bottom_m, bottom_b = find_line_through_2_points(back_right, front_right)
            bottom_IP_x = -bottom_b / bottom_m

        elif orientation == 'FR':
            top_m, top_b = find_line_through_2_points(front_right, back_right)
            top_IP_x = -top_b / top_m
            bottom_m, bottom_b = find_line_through_2_points(front_right, front_left)
            bottom_IP_x = -bottom_b / bottom_m
        else:
            raise('Correct frontal edge could not be determined after rotation!')

        if top_IP_x >= bottom_IP_x:
            transloc_front_left = (front_left[0] - top_IP_x + distance, front_left[1], 'FL')
            transloc_front_right = (front_right[0] - top_IP_x + distance, front_right[1], 'FR')
            transloc_back_left = (back_left[0] - top_IP_x + distance, back_left[1], 'BL')
            transloc_back_right = (back_right[0] - top_IP_x + distance, back_right[1], 'BR')
        else:
            transloc_front_left = (front_left[0] - bottom_IP_x + distance, front_left[1], 'FL')
            transloc_front_right = (front_right[0] - bottom_IP_x + distance, front_right[1], 'FR')
            transloc_back_left = (back_left[0] - bottom_IP_x + distance, back_left[1], 'BL')
            transloc_back_right = (back_right[0] - bottom_IP_x + distance, back_right[1], 'BR')

    vehicle.orientation = orientation
    vehicle.edge_points = (transloc_front_left, transloc_front_right, transloc_back_right, transloc_back_left)

    """
    # NOTE: This solution rotates the vehicle, gets the point with the shortest distance and locates this point at 
        coordinate 40/0. Exceptions are rotations of 0, 90, 180 and 270°. Here, the side of the vehicle is set to 40/0

    # calculate coordinates of the rotated vehicle centered at (0/0)
    front_left = (-length/2*math.cos(rad_rotation_angle)-broadness/2*math.sin(rad_rotation_angle),
                        length/2*math.sin(rad_rotation_angle)-broadness/2*math.cos(rad_rotation_angle))
    front_right = (-length/2*math.cos(rad_rotation_angle)+broadness/2*math.sin(rad_rotation_angle),
                        length/2*math.sin(rad_rotation_angle)+broadness/2*math.cos(rad_rotation_angle))
    back_left = (length/2*math.cos(rad_rotation_angle)-broadness/2*math.sin(rad_rotation_angle),
                        -length/2*math.sin(rad_rotation_angle)-broadness/2*math.cos(rad_rotation_angle))
    back_right = (length/2*math.cos(rad_rotation_angle)+broadness/2*math.sin(rad_rotation_angle),
                        -length/2*math.sin(rad_rotation_angle)+broadness/2*math.cos(rad_rotation_angle))


    edge_points = (front_right, front_left, back_left, back_right)
    sort = sorted(edge_points, key= lambda point: point[0])  # find the point most to the left

    min_x = sort[0][0]
    min_y = sort[0][1]


    # translocate all points along x-axis so that the most left point is at set distance
    if rotation_angle % 90 == 0:
        transloc_front_left = (front_left[0] - min_x + distance, front_left[1])
        transloc_front_right = (front_right[0] - min_x + distance, front_right[1])
        transloc_back_left = (back_left[0] - min_x + distance, back_left[1])
        transloc_back_right = (back_right[0] - min_x + distance, back_right[1])
    else:
        transloc_front_left = (front_left[0] - min_x + distance, front_left[1]-min_y)
        transloc_front_right = (front_right[0] - min_x + distance, front_right[1]-min_y)
        transloc_back_left = (back_left[0] - min_x + distance, back_left[1]-min_y)
        transloc_back_right = (back_right[0] - min_x + distance, back_right[1]-min_y)"""


def generate_shot(shooter, target, distance):
    """Checks if shot is natural hit or if scatter shot has to be calculated. If scatter, scatter function is called.
    Returns the shot coordinates and type (1 = natural, 0 = scatter)"""
    accuracy = shooter.calc_base_acc(distance) * target.target_size
    hit_dice = random.uniform(0,1)

    shooter.shot_counter += 1

    if accuracy >= 1:
        shot = (1000, 0)
        shot_type = 1
        shooter.natural_hit_counter += 1
    elif hit_dice <= accuracy:
        shot = (1000, 0)
        shot_type = 1
        shooter.natural_hit_counter += 1
    else:
        shot = generate_scatter_shot(shooter, distance)
        shot_type = 0
        shooter.scatter_shot_counter += 1

    return shot, shot_type


def generate_scatter_shot(shooter, distance):
    """Generates the coordinates and angle of a scattered shot."""

    scatter_offset = shooter.scatter_offset
    scatter_ratio = shooter.scatter_ratio
    distance_scatter_max = shooter.distance_scatter_max
    scatter_angle = shooter.scatter_angle

    # calculate longitudinal scatter
    longitudinal_scatter = min(distance_scatter_max, distance * scatter_ratio)
    scatter_centre = distance + scatter_offset * longitudinal_scatter
    longitudinal_scatter_min = scatter_centre - longitudinal_scatter
    longitudinal_scatter_max = scatter_centre + longitudinal_scatter
    scatter_shot_x = random.uniform(longitudinal_scatter_min, longitudinal_scatter_max)

    # calculate left/right scatter
    effective_angle = scatter_angle/2  # halving is necessary due to instructions
    angle_scatter = math.tan(math.radians(effective_angle)) * scatter_shot_x
    angle_scatter_min = -angle_scatter
    angle_scatter_max = angle_scatter
    scatter_shot_y = random.uniform(angle_scatter_min, angle_scatter_max)

    # get the randomized scatter shot
    scatter_shot = (scatter_shot_x, scatter_shot_y)

    return scatter_shot


def determine_hit(target, shot_coordinates):
    """Calculates if a shot was a hit or not. Also determines if front or rear armor was hit. Returns hit (False/True)
    and armor ('rear'/'front')"""
    # Logic:
    # 1. Takes orientation and edge coordinates of the target and calculates the angles
    # 2. Checks if the shot angle could potentially hit the target and checks if front or side armor were to be hit
    # 3. Calculates intersection point of the seide of the vehicle with the shot and tests if the shot flies far enough
    #           to hit the vehicle


    hit = False
    armor = None
    IP_dis = 0

    orientation = target.orientation
    front_left, front_right, back_right, back_left = target.edge_points

    front_left_angle = math.degrees(math.atan2(front_left[1], front_left[0]))
    front_right_angle = math.degrees(math.atan2(front_right[1], front_right[0]))
    back_right_angle = math.degrees(math.atan2(back_right[1], back_right[0]))
    back_left_angle = math.degrees(math.atan2(back_left[1], back_left[0]))

    shot_coordinate_x, shot_coordinate_y = shot_coordinates
    shot_angle = math.degrees(math.atan2(shot_coordinate_y, shot_coordinate_x))
    shot_dis = math.sqrt(shot_coordinate_x**2+shot_coordinate_y**2)


    if orientation == 'FL':
        if shot_angle >= front_left_angle and shot_angle <= front_right_angle:
            armor = "front"
            IP, IP_dis = calculate_lin_intersection(front_left, front_right, shot_angle)
        elif shot_angle < front_left_angle and shot_angle >= back_left_angle:
            IP, IP_dis = calculate_lin_intersection(front_left, back_left, shot_angle)
            dis_from_front = math.sqrt((IP[0]-front_left[0])**2+(IP[1]-front_left[1])**2)  # first half of the vehicle is frontal armor. If IP is closer to the front then use frontal armor.
            if dis_from_front < target.hitbox_length/2:
                armor = "front"
            elif dis_from_front == target.hitbox_length / 2:  # If shot hits the exactly the middle use 50% chance (important for +/- 90°)
                armor = random.choice(["rear", "front"])
            else:
                armor = "rear"
        else:  # shot has missed the target, hit = False is returned
            return hit, armor

    elif orientation == 'FR':
        if shot_angle >= front_right_angle and shot_angle <= back_right_angle:
            IP, IP_dis = calculate_lin_intersection(front_right, back_right, shot_angle)
            dis_from_front = math.sqrt((IP[0]-front_right[0])**2+(IP[1]-front_right[1])**2)  # first half of the vehicle is frontal armor. If IP is closer to the front then use frontal armor
            if dis_from_front < target.hitbox_length/2:
                armor = "front"
            elif dis_from_front == target.hitbox_length / 2:  # If shot hits the exactly the middle use 50% chance (important for +/- 90°)
                armor = random.choice(["rear", "front"])
            else:
                armor = "rear"
        elif shot_angle < front_right_angle and shot_angle >= front_left_angle:
            armor = "front"
            IP, IP_dis = calculate_lin_intersection(front_right, front_left, shot_angle)
        else:  # shot has missed the target, hit = False is returned
            return hit, armor

    elif orientation == 'BR':
        if shot_angle >= back_right_angle and shot_angle <= back_left_angle:
            IP, IP_dis = calculate_lin_intersection(back_right, back_left, shot_angle)
            armor = "rear"
        elif shot_angle < back_right_angle and shot_angle >= front_right_angle:
            IP, IP_dis = calculate_lin_intersection(back_right, front_right, shot_angle)
            dis_from_front = math.sqrt((IP[0]-front_right[0])**2+(IP[1]-front_right[1])**2)  # first half of the vehicle is frontal armor. If IP is closer to the front then use frontal armor
            if dis_from_front <= target.hitbox_length/2:
                armor = "front"
            elif dis_from_front == target.hitbox_length / 2:  # If shot hits the exactly the middle use 50% chance (important for +/- 90°)
                armor = random.choice(["rear", "front"])
            else:
                armor = "rear"
        else:  # shot has missed the target, hit = False is returned
            return hit, armor

    elif orientation == 'BL':
        if shot_angle >= back_left_angle and shot_angle <= front_left_angle:
            IP, IP_dis = calculate_lin_intersection(back_left, front_left, shot_angle)
            dis_from_front = math.sqrt((IP[0]-front_left[0])**2+(IP[1]-front_left[1])**2)  # first half of the vehicle is frontal armor. If IP is closer to the front then use frontal armor
            if dis_from_front < target.hitbox_length/2:
                armor = "front"
            elif dis_from_front == target.hitbox_length / 2:  # If shot hits the exactly the middle use 50% chance (important for +/- 90°)
                armor = random.choice(["rear", "front"])
            else:
                armor = "rear"
        elif shot_angle < back_left_angle and shot_angle >= back_right_angle:
            armor = "rear"
            IP, IP_dis = calculate_lin_intersection(back_left, back_right, shot_angle)
        else:  # shot has missed the target, hit = False is returned
            return hit, armor


    if shot_dis >= IP_dis:
        hit = True

    return hit, armor


def determine_penetrating_hit(shooter, target, distance, armor_type):
    """Determines penetration and armor and applies damage if shot penetrates.
    Returns shot_type (1 = penetrated, 0 = bounce)"""

    pen = shooter.calc_pen(distance)
    if armor_type == "front":
        armor = target.armor
    else:
        armor = target.rear_armor

    pen_chance = pen/armor
    pen_dice = random.uniform(0, 1)

    if pen_chance >= 1:
        target.hitpoints -= shooter.damage  # apply damage to the target
        hit_type = 1
        shooter.pen_counter += 1
    elif pen_dice <= pen_chance:
        target.hitpoints -= shooter.damage  # apply damage to the target
        hit_type = 1
        shooter.pen_counter += 1
    else:
        target.hitpoints -= shooter.damage * shooter.deflection_dmg_modifier  # apply deflection damage to target
        hit_type = 0
        shooter.bounce_counter += 1

    return hit_type


def calculate_lin_intersection(edge1, edge2, shot_angle):
    """Calculates intersection of two lines. Needed to test if the tank gets hit by a scatter shot or not."""

    # Calc slope of the shot
    m_1 = math.tan(math.radians(shot_angle))

    try:
        # Calc line of the tank hull
        m_2, b_2 = find_line_through_2_points(edge1, edge2)

        # Calc intersection point
        IP_x = b_2/(m_1-m_2)
        IP_y = m_1 * IP_x
        IP = (IP_x, IP_y)

    except ZeroDivisionError:  # If tank hull was rotated to 0, 90, 180 or 270 degrees
        IP_x = edge1[0]
        IP_y = m_1 * IP_x
        IP = (IP_x, IP_y)

    IP_dis = math.sqrt((IP_x**2+IP_y**2))

    return IP, IP_dis


def find_line_through_2_points(point1, point2):
    """Calculates slope and intersection with Y-axis of a line from two points"""
    point1_x, point1_y, *unimp = point1
    point2_x, point2_y, *unimp = point2

    m = (point1_y-point2_y) / (point1_x-point2_x)
    b = point1_y - (m * point1_x)

    return m, b


def determine_hits_to_kill(shooter, target):
    hits_to_kill = math.ceil(target.hitpoints/shooter.damage)
    return hits_to_kill


def check_win_conditions(vehicle1, vehicle2):
    """Checks if one vehicle died"""

    win_condition = False

    winner = "None"
    looser = "None"

    if vehicle1.hitpoints <= 0 < vehicle2.hitpoints:
        win_condition = True
        winner = vehicle2
        looser = vehicle1
    elif vehicle2.hitpoints <= 0 < vehicle1.hitpoints:
        win_condition = True
        winner = vehicle1
        looser = vehicle2

    return win_condition, winner, looser


def initialize_targeting_output_files():
    with open("targeting_summary.txt", "w") as file:
        file.write("shooter; target; simulation_no; no_of_shots; time_passed; natural_shots; scatter_shots; total_hits; "
                    "scatter_hits; misses; no_of_pens; no_of_bounces; no_of_frontal_hits; no_of_rear_hits")
        file.write("\n")

    with open("detailed_targeting_report.txt", "w") as file:
        file.write("shooter; target; simulation_no; no_of_shot; no_of_damages; natural_shot; scatter_shot; hit; "
                    "scatter_hit; miss; pen; bounce; frontal_hit; rear_hit")
        file.write("\n")


def sim_tank_targeting(no_of_iterations, chosen_shooter, chosen_target, distance, target_angle):
    """Simulates one tank shooting at another for x interations. Good for finding out about real hit rates due to
    scatter and bounce rates due to angling."""
    print("Calculating...")
    initialize_targeting_output_files()
    simulation_no = 0

    for i in range(no_of_iterations):

        shooter = copy.copy(unit_dict[chosen_shooter])
        target = copy.copy(unit_dict[chosen_target])

        shooter.calc_time_to_next_shot()

        translocate_vehicle(target, distance, target_angle)

        simulation_no += 1

        no_of_shots = 0
        natural_hits = 0
        real_hit_counter = 0
        pen_counter = 0
        frontal_hits_counter = 0
        rear_hits_counter = 0

        time_passed = shooter.fire_aim_time

        while target.hitpoints > 0:
            no_of_shots += 1
            shot, shot_type = generate_shot(shooter, target, distance)
            natural_hits += shot_type

            it_frontal_hit = 0
            it_rear_hit = 0

            time_passed += shooter.time_to_next_shot
            shooter.calc_time_to_next_shot()

            hit, armor = determine_hit(target, shot)
            real_hit_counter += hit
            if hit == True and armor == "front":
                frontal_hits_counter += 1
                it_frontal_hit = 1
            elif hit == True and armor == "rear":
                rear_hits_counter += 1
                it_rear_hit = 1

            if hit:
                hit_type = determine_penetrating_hit(shooter, target, distance, armor)
                pen_counter += hit_type
            else:
                hit_type = 0

            it_no_of_shots = no_of_shots
            no_of_damages = pen_counter
            it_natural_shot = shot_type
            it_scatter_shot = 1 - shot_type
            it_hit = hit
            it_scatter_hit = hit - shot_type
            it_miss = 1 - hit
            it_pen = hit_type
            it_bounce = hit - hit_type

            iteration_stats = (
            simulation_no, it_no_of_shots, no_of_damages, it_natural_shot, it_scatter_shot, it_hit, it_scatter_hit,
            it_miss, it_pen, it_bounce, it_frontal_hit, it_rear_hit)

            report_targeting_iteration(shooter, target, iteration_stats)

        no_of_shots = no_of_shots
        natural_shots = natural_hits
        scatter_shots = no_of_shots - natural_hits
        total_hits = real_hit_counter
        scatter_hits = real_hit_counter - natural_hits
        misses = no_of_shots - real_hit_counter
        no_of_pens = pen_counter
        no_of_bounces = real_hit_counter - pen_counter
        no_of_frontal_hits = frontal_hits_counter
        no_of_rear_hits = rear_hits_counter

        processed_time_passed = str(time_passed).replace(".", ",")   # needed for Excel because of decimal separator

        summary_stats = (
        simulation_no, no_of_shots, processed_time_passed, natural_shots, scatter_shots, total_hits, scatter_hits, misses, no_of_pens,
        no_of_bounces, no_of_frontal_hits, no_of_rear_hits)
        report_targeting_summary(shooter, target, summary_stats)

    print("Simulation complete. Output files written.")
    print()


def report_targeting_summary(shooter, target, combat_stats):

    with open("targeting_summary.txt", "a") as file:
        file.write(shooter.name)
        file.write(";")
        file.write(target.name)
        file.write(";")
        for element in combat_stats:
            file.write(str(element))
            file.write(";")
        file.write("\n")


def report_targeting_iteration(shooter, target, iteration_stats):
    with open("detailed_targeting_report.txt", "a") as file:
        file.write(shooter.name)
        file.write(";")
        file.write(target.name)
        file.write(";")
        for element in iteration_stats:
            file.write(str(element))
            file.write(",")
        file.write("\n")


def initialize_tank_sim_output_files():
    with open("tank_sim_summary.txt", "w") as file:
        file.write("simulation_no; winner; looser; time; round_no; winner_health; winner_no_of_shots; winner_scatter_shots; "
           "winner_total_hits; winner_natural_hits; winner_misses; winner_pens; winner_bounces; looser_health; "
           "looser_no_of_shots; looser_scatter_shots; looser_total_hits; looser_natural_hits; "
           "looser_misses; looser_pens; looser_bounces")

    #with open("fight_report", "w") as fi_rep:
    #    fi_rep.write("shooter, target, simulation_no, no_of_shot, no_of_damages, natural_shot, scatter_shot, hit, "
    #                "scatter_hit, miss, pen, bounce, frontal_hit, rear_hit")


def determine_next_shooter(vehicle1, vehicle2, double_shot_cutoff=0.1):
    """Checks who has the next shot and if the time to next shot of the target (double_shot_cutoff)
     is lower than default=0.1 seconds (this allows the target to shoot back if true). Then calculates the time
     to the next shot for the next iteration for both vehicles. Returns shooter, target, double_shot."""

    double_shot = False

    if abs(vehicle1.time_to_next_shot - vehicle2.time_to_next_shot) < double_shot_cutoff:
        double_shot = True

    if vehicle1.time_to_next_shot <= vehicle2.time_to_next_shot:
        shooter = vehicle1
        target = vehicle2
        passed_time = vehicle1.time_to_next_shot
        vehicle2.time_to_next_shot -= vehicle1.time_to_next_shot
        vehicle1.calc_time_to_next_shot()
    elif vehicle1.time_to_next_shot > vehicle2.time_to_next_shot:
        shooter = vehicle2
        target = vehicle1
        passed_time = vehicle2.time_to_next_shot
        vehicle1.time_to_next_shot -= vehicle2.time_to_next_shot
        vehicle2.calc_time_to_next_shot()


    return shooter, target, double_shot, passed_time


def sim_tank_fight(chosen_vehicle1, vehicle1_rot_angle, chosen_vehicle2, vehicle2_rot_angle, vehicle2_shot_delay, distance,
                   no_of_sims, retal_shot_time=0.1):
    """Simulates two tanks shooting at each other at given angles and distance until one is dead. Reports winner, looser
    and several combat stats"""


    print("Calculating...")
    initialize_tank_sim_output_files()


    for sim_no in range(1, no_of_sims+1):

        vehicle1 = copy.copy(unit_dict[chosen_vehicle1])
        vehicle2 = copy.copy(unit_dict[chosen_vehicle2])

        translocate_vehicle(vehicle1, distance, vehicle1_rot_angle)
        translocate_vehicle(vehicle2, distance, vehicle2_rot_angle)

        vehicle1.time_to_next_shot += vehicle1.fire_aim_time
        vehicle2.time_to_next_shot += vehicle2.fire_aim_time + vehicle2_shot_delay

        fight_ended = False
        no_of_rounds = 0
        fight_duration = 0

        while not fight_ended:
            no_of_rounds += 1

            shooter, target, retaliation_shot, passed_time = determine_next_shooter(vehicle1, vehicle2, retal_shot_time)

            fight_duration += passed_time

            shot, shot_type = generate_shot(shooter, target, distance)
            hit, armor = determine_hit(target, shot)
            if hit:
                shooter.hit_counter += 1
                hit_type = determine_penetrating_hit(shooter, target, distance, armor)
            else:
                hit_type = 0
                shooter.miss_counter += 1

            fight_ended, winner, looser = check_win_conditions(shooter, target)

            if fight_ended:  # checks if the loser can shoot back and then ends the sim.
                if retaliation_shot:
                    shot, shot_type = generate_shot(target, shooter, distance)
                    hit, armor = determine_hit(shooter, shot)
                    if hit:
                        hit_type = determine_penetrating_hit(target, shooter, distance, armor)
                        target.hit_counter += 1
                    else:
                        target.miss_counter += 1


        if winner.hitpoints <= 0 and looser.hitpoints <= 0:
            winner.name = "DRAW " + winner.name
            looser.name = "DRAW " + looser.name


        combat_summary = (sim_no, winner.name, looser.name, fight_duration, no_of_rounds, winner.hitpoints, winner.shot_counter,
                          winner.scatter_shot_counter, winner.hit_counter, winner.natural_hit_counter,
                          winner.miss_counter, winner.pen_counter, winner.bounce_counter, looser.hitpoints,
                          looser.shot_counter, looser.scatter_shot_counter, looser.hit_counter,
                          looser.natural_hit_counter, looser.miss_counter, looser.pen_counter, looser.bounce_counter)

        report_tank_sim_summary(combat_summary)

    print("Simulation complete. Output file written.")
    print()


def report_tank_sim_summary(combat_summary):
    with open("tank_sim_summary.txt", "a") as file:
        file.write("\n")
        for element in combat_summary:
            file.write(str(element))
            file.write(";")


def user_main_menu():
    exit_programm = False
    while not exit_programm:
        user_inp = input("Choose mode (type number): [1] inspect, [2] tank targeting simulator, "
                         "[3] tank fight simulator, [4] exit  ")
        if user_inp == "1":
            user_inspect()
        elif user_inp == "2":
            it, sho, tar, dis, an = get_user_input_tank_targeting()
            sim_tank_targeting(it, sho, tar, dis, an)
        elif user_inp == "3":
            v1, v1_rot, v2, v2_rot, v2_del, dis, it, retal_time = get_user_input_tank_sim()
            sim_tank_fight(v1, v1_rot, v2, v2_rot, v2_del, dis, it, retal_time)
        elif user_inp == "4":
            exit_programm = True


def user_inspect():
    for key, value in unit_dict.items():
        print(key)
    print()
    choosen_vehicle_inp = ""

    while choosen_vehicle_inp not in unit_dict:
        try:
            choosen_vehicle_inp = input("Choose a vehicle  ")
            choosen_vehicle = unit_dict[choosen_vehicle_inp]
            vehicle_attr = vars(choosen_vehicle)
            print(vehicle_attr)
            print()
        except KeyError:
            print("Vehicle not known. Try again.")


def get_user_input_tank_sim():
    chosen_vehicle1 = None
    chosen_vehicle1_angle = None
    chosen_vehicle2 = None
    chosen_vehicle2_angle = None
    chosen_vehicle2_shot_delay = None
    chosen_distance = None
    chosen_no_of_iterations = None
    chosen_retaliation_shot_time = None

    for key, value in unit_dict.items():
        print(key)
    print()

    while chosen_vehicle1 not in unit_dict.keys():
        try:
            chosen_vehicle1 = input("Choose vehicle 1  ")
        except KeyError:
            print("Vehicle not in unit list")

    while not isinstance(chosen_vehicle1_angle, float):
        try:
            chosen_vehicle1_angle = float(input("Choose vehicle 1 rotation angle  "))
        except ValueError:
            print("Angle must be number")

    while chosen_vehicle2 not in unit_dict.keys():
        try:
            chosen_vehicle2 = input("Choose vehicle 2  ")
        except KeyError:
            print("Vehicle not in unit list")

    while not isinstance(chosen_vehicle2_angle, float):
        try:
            chosen_vehicle2_angle = float(input("Choose vehicle 2 rotation angle  "))
        except ValueError:
            print("Angle must be number")

    while not isinstance(chosen_vehicle2_shot_delay, float):
        try:
            chosen_vehicle2_shot_delay = float(input("Choose vehicle 2 shot delay  "))
        except ValueError:
            print("Delay must be number")

    while not isinstance(chosen_distance, float):
        try:
            chosen_distance = float(input("Choose distance  "))
        except ValueError:
            print("Distance must be number")

    while not isinstance(chosen_no_of_iterations, int):
        try:
            chosen_no_of_iterations = int(input("Choose number of iterations  "))
        except ValueError:
            print("Number of iterations must be integer")

    while not isinstance(chosen_retaliation_shot_time, float):
        try:
            chosen_retaliation_shot_time = float(input("Choose retaliation shot time (default = 0.1 s)  "))
        except ValueError:
            chosen_retaliation_shot_time = 0.1
            print("Defaulted to 0.1 s")

    return chosen_vehicle1, chosen_vehicle1_angle, chosen_vehicle2, chosen_vehicle2_angle, chosen_vehicle2_shot_delay, \
           chosen_distance, chosen_no_of_iterations, chosen_retaliation_shot_time


def get_user_input_tank_targeting():
    chosen_shooter = None
    chosen_target = None
    chosen_target_angle = None
    chosen_distance = None
    chosen_no_of_iterations = None

    for key, value in unit_dict.items():
        print(key)
    print()

    while chosen_shooter not in unit_dict.keys():
        try:
            chosen_shooter = input("Choose shooter  ")
        except KeyError:
            print("Vehicle not in unit list")

    while chosen_target not in unit_dict.keys():
        try:
            chosen_target = input("Choose target  ")
        except KeyError:
            print("Vehicle not in unit list")

    while not isinstance(chosen_target_angle, float):
        try:
            chosen_target_angle = float(input("Choose target angle  "))
        except ValueError:
            print("Angle must be number")

    while not isinstance(chosen_distance, float):
        try:
            chosen_distance = float(input("Choose distance  "))
        except ValueError:
            print("Distance must be number")

    while not isinstance(chosen_no_of_iterations, int):
        try:
            chosen_no_of_iterations = int(input("Choose number of iterations  "))
        except ValueError:
            print("Number of iterations must be integer")

    return chosen_no_of_iterations, chosen_shooter, chosen_target, chosen_distance, chosen_target_angle



    #while
    #vehicle1 = unit_dict["input"]


    pass



unit_dict = build_unit_dictionary()

print("DeCoHde - Snake Edition!")
print("v1.0")
print()

user_main_menu()






